@@ -138,6 +138,8 @@ static inline void display_help(char *str){
   fprintf(stdout, "         compute the saturation of the ideal\n");
   fprintf(stdout, "         generated by the first k-1 polynomials\n");
   fprintf(stdout, "         with respect to the kth polynomial.\n");
+  fprintf(stdout, "         Note: At the moment restricted to 32 bit\n");
+  fprintf(stdout, "         prime fields.\n");
   fprintf(stdout, "-u UHT   Number of steps after which the\n");
   fprintf(stdout, "         hash table is newly generated.\n");
   fprintf(stdout, "         Default: 0, i.e. no update.\n");
@@ -409,7 +411,7 @@ int main(int argc, char **argv){
       fclose(ofile);
     }
     /**
-       We get from files the requested data. 
+       We get from files the requested data.
     **/
     //  int32_t mon_order   = 0;
     int32_t nr_vars     = 0;
@@ -433,7 +435,7 @@ int main(int argc, char **argv){
         }
         la_option = 2;
     }
-    
+
     /* data structures for parametrization */
     param_t *param  = NULL;
     mpz_param_t *mpz_paramp = malloc(sizeof(mpz_param_t));
@@ -469,7 +471,7 @@ int main(int argc, char **argv){
     if (info_level > 0) {
         double st1 = cputime();
         double rt1 = realtime();
-        fprintf(stderr, "-------------------------------------------------\
+        fprintf(stderr, "\n\n-------------------------------------------------\
 -----------------------------------\n");
         fprintf(stderr, "msolve overall time  %13.2f sec (elapsed) / %5.2f sec (cpu)\n",
                 rt1-rt0, st1-st0);
