@@ -98,8 +98,6 @@ void free_f4_julia_result_data(
          for (i = 0; i < len; ++i) {
              mpz_clear((*cfs)[i]);
          }
-         (*freep)(*cfs);
-         *cfs  = NULL; 
     } else {
         if (field_char > 0) {
             int32_t *cfs  = *((int32_t **)bcf);
