@@ -1837,10 +1837,11 @@ int64_t f4_trace_julia(
     int *invalid_gens       =   NULL;
     int32_t use_signatures  =   0;
     int32_t nr_nf           =   0;
+    int32_t truncate_lifting =  0;
     int res = validate_input_data(&invalid_gens, cfs, lens, &field_char, &mon_order,
             &elim_block_len, &nr_vars, &nr_gens, &nr_nf, &ht_size, &nr_threads,
             &max_nr_pairs, &reset_ht, &la_option, &use_signatures, &reduce_gb,
-            &info_level);
+            &truncate_lifting, &info_level);
 
     /* all data is corrupt */
     if (res == -1) {
@@ -1854,7 +1855,7 @@ int64_t f4_trace_julia(
                 field_char, mon_order, elim_block_len, nr_vars, nr_gens,
                 nr_nf, ht_size, nr_threads, max_nr_pairs, reset_ht, la_option,
                 use_signatures, reduce_gb, prime_start, nr_primes, pbm_file,
-                info_level)) {
+                truncate_lifting, info_level)) {
         return 0;
     }
 
